@@ -68,6 +68,7 @@ pub trait VhostUserBackendReqHandler {
     fn set_config(&self, offset: u32, buf: &[u8], flags: VhostUserConfigFlags) -> Result<()>;
     fn set_backend_req_fd(&self, _backend: Backend) {}
     fn set_gpu_socket(&self, _gpu_backend: GpuBackend) -> Result<()>;
+    fn get_shared_object(&self, uuid: VhostUserSharedMsg) -> Result<File>;
     fn get_inflight_fd(&self, inflight: &VhostUserInflight) -> Result<(VhostUserInflight, File)>;
     fn set_inflight_fd(&self, inflight: &VhostUserInflight, file: File) -> Result<()>;
     fn get_max_mem_slots(&self) -> Result<u64>;
@@ -129,6 +130,7 @@ pub trait VhostUserBackendReqHandlerMut {
     fn set_config(&mut self, offset: u32, buf: &[u8], flags: VhostUserConfigFlags) -> Result<()>;
     fn set_backend_req_fd(&mut self, _backend: Backend) {}
     fn set_gpu_socket(&mut self, _gpu_backend: GpuBackend) -> Result<()>;
+    fn get_shared_object(&mut self, uuid: VhostUserSharedMsg) -> Result<File>;
     fn get_inflight_fd(
         &mut self,
         inflight: &VhostUserInflight,
@@ -248,6 +250,10 @@ impl<T: VhostUserBackendReqHandlerMut> VhostUserBackendReqHandler for Mutex<T> {
         self.lock().unwrap().set_gpu_socket(gpu_backend)
     }
 
+    fn get_shared_object(&self, uuid: VhostUserSharedMsg) -> Result<File> {
+        self.lock().unwrap().get_shared_object(uuid)
+    }
+
     fn get_inflight_fd(&self, inflight: &VhostUserInflight) -> Result<(VhostUserInflight, File)> {
         self.lock().unwrap().get_inflight_fd(inflight)
     }
@@ -321,7 +327,6 @@ pub struct BackendReqHandler<S: VhostUserBackendReqHandler> {
 
     virtio_features: u64,
     acked_virtio_features: u64,
-    protocol_features: VhostUserProtocolFeatures,
     acked_protocol_features: u64,
 
     // sending ack for messages without payload
@@ -341,7 +346,6 @@ impl<S: VhostUserBackendReqHandler> BackendReqHandler<S> {
             backend,
             virtio_features: 0,
             acked_virtio_features: 0,
-            protocol_features: VhostUserProtocolFeatures::empty(),
             acked_protocol_features: 0,
             reply_ack_enabled: false,
             error: None,
@@ -506,7 +510,9 @@ impl<S: VhostUserBackendReqHandler> BackendReqHandler<S> {
             }
             Ok(FrontendReq::GET_PROTOCOL_FEATURES) => {
                 self.check_request_size(&hdr, size, 0)?;
-                let features = self.backend.get_protocol_features()?;
+                // REPLY_ACK is implemented entirely by this library, so it's always supported.
+                let features =
+                    self.backend.get_protocol_features()? | VhostUserProtocolFeatures::REPLY_ACK;
 
                 // Enable the `XEN_MMAP` protocol feature for backends if xen feature is enabled.
                 #[cfg(feature = "xen")]
@@ -514,7 +520,6 @@ impl<S: VhostUserBackendReqHandler> BackendReqHandler<S> {
 
                 let msg = VhostUserU64::new(features.bits());
                 self.send_reply_message(&hdr, &msg)?;
-                self.protocol_features = features;
                 self.update_reply_ack_flag();
             }
             Ok(FrontendReq::SET_PROTOCOL_FEATURES) => {
@@ -563,6 +568,26 @@ impl<S: VhostUserBackendReqHandler> BackendReqHandler<S> {
                 let res = self.set_backend_req_fd(files);
                 self.send_ack_message(&hdr, res)?;
             }
+            Ok(FrontendReq::GET_SHARED_OBJECT) => {
+                self.check_proto_feature(VhostUserProtocolFeatures::SHARED_OBJECT)?;
+                self.check_request_size(&hdr, size, hdr.get_size() as usize)?;
+                let msg = self.extract_request_body::<VhostUserSharedMsg>(&hdr, size, &buf)?;
+
+                let reply_hdr = self.new_reply_header::<VhostUserEmpty>(&hdr, 0)?;
+                match self.backend.get_shared_object(msg) {
+                    Ok(file) => {
+                        self.main_sock.send_message(
+                            &reply_hdr,
+                            &VhostUserEmpty,
+                            Some(&[file.as_raw_fd()]),
+                        )?;
+                    }
+                    Err(_) => {
+                        self.main_sock
+                            .send_message(&reply_hdr, &VhostUserEmpty, None)?;
+                    }
+                }
+            }
             Ok(FrontendReq::GET_INFLIGHT_FD) => {
                 self.check_proto_feature(VhostUserProtocolFeatures::INFLIGHT_SHMFD)?;
 
@@ -658,17 +683,17 @@ impl<S: VhostUserBackendReqHandler> BackendReqHandler<S> {
             Ok(FrontendReq::POSTCOPY_ADVISE) => {
                 self.check_proto_feature(VhostUserProtocolFeatures::PAGEFAULT)?;
 
+                let reply_hdr = self.new_reply_header::<VhostUserEmpty>(&hdr, 0)?;
                 let res = self.backend.postcopy_advice();
                 match res {
-                    Ok(uffd_file) => {
-                        let hdr = self.new_reply_header::<VhostUserEmpty>(&hdr, 0)?;
-                        self.main_sock.send_message(
-                            &hdr,
-                            &VhostUserEmpty,
-                            Some(&[uffd_file.as_raw_fd()]),
-                        )?
-                    }
-                    Err(_) => self.main_sock.send_message(&hdr, &VhostUserEmpty, None)?,
+                    Ok(uffd_file) => self.main_sock.send_message(
+                        &reply_hdr,
+                        &VhostUserEmpty,
+                        Some(&[uffd_file.as_raw_fd()]),
+                    )?,
+                    Err(_) => self
+                        .main_sock
+                        .send_message(&reply_hdr, &VhostUserEmpty, None)?,
                 }
             }
             #[cfg(feature = "postcopy")]
@@ -926,7 +951,6 @@ impl<S: VhostUserBackendReqHandler> BackendReqHandler<S> {
         let pflag = VhostUserProtocolFeatures::REPLY_ACK;
 
         self.reply_ack_enabled = (self.virtio_features & vflag) != 0
-            && self.protocol_features.contains(pflag)
             && (self.acked_protocol_features & pflag.bits()) != 0;
     }
 
